#!/usr/bin/env python

import os
import os.path
import re
import subprocess
import sys

from Bio import SearchIO, SeqIO
import numpy as np
from tqdm import tqdm

from .args_processing import process_args
from .parser import create_parser
from .helper import (
    create_directories, read_input_files, 
    conduct_hmm_search, set_names,
    handle_single_copy_writing, handle_multi_copy_writing,
    handle_absent_writing, handle_percent_present_summary,
    check_hmmsearch_output
    )

def execute(
    fasta_file_list: str,
    hmms_file_list: str,
    evalue: float,
    percent_bitscore: float,
    output_dir: str,
    cpu: int
):

    # read input files
    fasta_file_list, hmms_file_list = read_input_files(fasta_file_list, hmms_file_list)

    # create directories that will be
    # populated with output files
    create_directories(output_dir)

    # create empty object for checking 
    # hmmsearches with no hits
    exhausted = object()

    ## loop through fasta files
    for fasta in tqdm(fasta_file_list):
        # ortholog presence absence stats will hold
        # idx0: # single-copy orthologs
        # idx1: # multi-copy orthologs
        # idx2: # absent orthologs
        ortholog_presence_absence_stats = [0, 0, 0]
        record_dict = SeqIO.to_dict(SeqIO.parse(fasta[0], "fasta"))
        fasta_name = re.sub(r'^.*/', '', fasta[0])

        ## loop through HMMs
        for hmm in hmms_file_list:
            hmm_name, \
            singlecopy_name, \
            all_name, \
            short_summary_name, \
            long_summary_name, \
            hmmsearch_out = set_names(hmm[0], fasta_name, output_dir)
            
            # execute hmmsearch
            conduct_hmm_search(hmmsearch_out, hmm[0], fasta[0], evalue, cpu)

            # identify single copy orthologs and write them to separate scog files
            check_hmmsearch_output(hmmsearch_out)

            with open(hmmsearch_out, 'r') as hmmsearch_out:
                have_written = 0
                for qresult in SearchIO.parse(hmmsearch_out, 'hmmer3-tab'):
                    matches = qresult.hits
                    top_score = matches[0].bitscore
                    hits = []

                    # remove genes with bitscores less than specified value (-b arg)
                    # of the bitscore value as the top hit in the hmmsearch
                    for hit in matches:
                        if hit.bitscore >= (percent_bitscore*top_score):
                            hits.append(hit)
                    
                    num_hits = len(hits)

                    # if there is no second column provided,
                    # create a placeholder name that is an empty string
                    try:
                        name = fasta[1] + " "
                    except IndexError:
                        name = ""

                    # if single copy
                    if num_hits == 1:
                        # write to single copy orthologous gene file
                        # write to all gene file
                        # write to long log file
                        handle_single_copy_writing(
                            singlecopy_name,
                            name,
                            all_name,
                            long_summary_name,
                            fasta_name,
                            hmm_name,
                            hits,
                            record_dict
                        )

                        # add to counter
                        ortholog_presence_absence_stats[0]+=1
                        have_written+=1

                    # if multi-copy
                    elif num_hits > 1:
                        handle_multi_copy_writing(
                            name,
                            all_name,
                            long_summary_name,
                            fasta_name,
                            hmm_name,
                            hits,
                            record_dict,
                            num_hits
                        )
                        
                        # add to counter
                        ortholog_presence_absence_stats[1]+=1
                        have_written+=1
                    
                # if absent
                if have_written == 0:
                    handle_absent_writing(
                        long_summary_name,
                        fasta_name,
                        hmm_name
                    )

                    # add to counter
                    ortholog_presence_absence_stats[2]+=1
        
        handle_percent_present_summary(
            ortholog_presence_absence_stats,
            fasta_name,
            short_summary_name
        )

def main(argv=None):
    """
    Function that parses and collects arguments              
    """
    # parse and assign arguments
    parser = create_parser()
    args = parser.parse_args()

    # pass to master execute function
    execute(**process_args(args))

if __name__ == "__main__":
    main(sys.argv[1:])



import os
import os.path
import re
import shutil
import subprocess
import sys

def read_input_files(
    fasta_file_list: str,
    hmms_file_list: str
):
    """
    read input files from user args
    """

    ## read input files
    fasta_file_list = [i.strip('\n').split('\t') for i in open(fasta_file_list)]
    hmms_file_list = [i.strip('\n').split('\t') for i in open(hmms_file_list)]

    return fasta_file_list, hmms_file_list

def create_directories(output_dir):
    """
    create directories for outputting files
    """

    ## create output directories
    try:
        os.mkdir(output_dir)
    except FileExistsError:
        print(f"WARNING: {output_dir} output exists. Overwriting directory now...")
        shutil.rmtree(output_dir)
        os.mkdir(output_dir)
    os.mkdir(f'{output_dir}/scog')
    os.mkdir(f'{output_dir}/hmmsearch_output')
    os.mkdir(f'{output_dir}/all_sequences')

def conduct_hmm_search(
    hmmsearch_out: str,
    hmm: str,
    fasta: str,
    evalue: float,
    cpu: int
):
    """
    execute hmm search using subprocess
    """
    # execute hmmsearch
    subprocess.call(
        [
            'hmmsearch', '--noali', 
            '--notextw', '-E', str(evalue),
            '--cpu', str(cpu),
            '--tblout', hmmsearch_out, 
            hmm, fasta
        ], stdout=subprocess.DEVNULL
    )

def check_hmmsearch_output(
    hmmsearch_out: str
):
    
    if not os.path.isfile(hmmsearch_out):
        print("HMM search failed. Check e-value is an appropriate number.")
        sys.exit()

def set_names(
    hmm: str,
    fasta_name: str,
    output_dir: str
):
    """
    set names for output files
    """
    hmm_name = re.sub(r'^.*/', '', hmm)
    singlecopy_name = f'{output_dir}/scog/{hmm_name}.orthofisher'
    all_name = f'{output_dir}/all_sequences/{hmm_name}.orthofisher'
    long_summary_name = f'{output_dir}/long_summary.txt'
    short_summary_name = f'{output_dir}/short_summary.txt'
    hmmsearch_out = f'{output_dir}/hmmsearch_output/{fasta_name}_{hmm_name}'

    return hmm_name, singlecopy_name, all_name, short_summary_name, long_summary_name, hmmsearch_out

def handle_single_copy_writing(
    singlecopy_name: str,
    fasta: str,
    all_name: str,
    long_summary_name: str,
    fasta_name: str,
    hmm_name: str,
    hits,
    record_dict
):
    """
    writes single copy entry to two output fasta files
    and to a long-form log file
    """

    
    entry_name = ''.join([fasta + hits[0]._id + '|' + str(hits[0].evalue) + '|' + str(hits[0].bitscore)])
    record_entry = record_dict[hits[0].id]

    # write single copy orthologous gene file
    with open(singlecopy_name, 'a') as f:
        f.write(">" + entry_name + '\n')
        sequence = [record_entry.seq[i:i+60] for i in range(0, len(record_entry.seq), 60)]
        for seq in sequence:
            if type(seq._data) is str:
                f.write(seq._data + '\n')
            else:
                f.write(seq._data.decode("utf-8") + '\n')

    # write to all gene file
    with open(all_name, 'a') as f:
        f.write(">" + entry_name + '\n')
        sequence = [record_entry.seq[i:i+60] for i in range(0, len(record_entry.seq), 60)]
        for seq in sequence:
            if type(seq._data) is str:
                f.write(seq._data + '\n')
            else:
                f.write(seq._data.decode("utf-8") + '\n')

    # write to long log file
    with open(long_summary_name, 'a') as f:
        f.write('\t'.join([fasta_name, hmm_name, 'single-copy', str(1), hits[0]._id]) + '\n')

def handle_multi_copy_writing(
    fasta: str,
    all_name: str,
    long_summary_name: str,
    fasta_name: str,
    hmm_name: str,
    hits,
    record_dict,
    num_hits: int
):
    """
    writes single copy entry to two output fasta files
    and to a long-form log file
    """

    hit_ids = []
    for idx in range(num_hits):
        hit_ids.append(hits[0]._id)

        entry_name = ''.join([fasta + hits[idx]._id + '|' + str(hits[idx].evalue) + '|' + str(hits[idx].bitscore)])
        record_entry = record_dict[hits[idx].id]

        # write to all gene file
        with open(all_name, 'a') as f:
            f.write(">" + entry_name + '\n')
            sequence = [record_entry.seq[i:i+60] for i in range(0, len(record_entry.seq), 60)]
            for seq in sequence:
                if type(seq._data) is str:
                    f.write(seq._data + '\n')
                else:
                    f.write(seq._data.decode("utf-8") + '\n')

        # write to long log file
        with open(long_summary_name, 'a') as f:
            f.write('\t'.join([fasta_name, hmm_name, 'multi-copy', str(num_hits), hits[idx]._id]) + '\n')

def handle_absent_writing(
    long_summary_name: str,
    fasta_name: str,
    hmm_name: str
):
    """
    logs if hmm is absent in the genome
    """
    # write to long log file
    with open(long_summary_name, 'a') as f:
        f.write('\t'.join([fasta_name, hmm_name, 'absent', str(0), 'NA']) + '\n')

def handle_percent_present_summary(
    ortholog_presence_absence_stats: list,
    fasta_name: str,
    short_summary_name: str
):
    """
    write log of percent present single-copy, multi-copy, or absent
    """

    header = [
        'file_name', 'single-copy', 'multi-copy', 'absent',
        'per_single-copy', 'per_multi-copy', 'per_absent'
    ]
    ortholog_presence_absence_per = []

    for val in ortholog_presence_absence_stats:
        ortholog_presence_absence_per.append(round(val/sum(ortholog_presence_absence_stats), 2))

    ortholog_presence_absence_stats = [str(i) for i in ortholog_presence_absence_stats]
    ortholog_presence_absence_per = [str(i) for i in ortholog_presence_absence_per] 

    if os.path.isfile(short_summary_name):
        with open(short_summary_name, 'a') as f:
            f.write(
                '\t'.join(
                    [fasta_name,
                    '\t'.join(ortholog_presence_absence_stats),
                    '\t'.join(ortholog_presence_absence_per)
                    ]
                ) + '\n'
            )
    else:
        with open(short_summary_name, 'a') as f:
            f.write('\t'.join(header) + '\n')

            f.write(
                '\t'.join(
                    [fasta_name,
                    '\t'.join(ortholog_presence_absence_stats),
                    '\t'.join(ortholog_presence_absence_per)
                    ]
                ) + '\n'
            )



import sys
import textwrap

from argparse import (
    ArgumentParser,
    RawTextHelpFormatter,
    SUPPRESS,
    RawDescriptionHelpFormatter,
)

from .version import __version__

def create_parser():
    parser = ArgumentParser(
        add_help=False,
        formatter_class=RawDescriptionHelpFormatter,
        usage=SUPPRESS,
        description=textwrap.dedent(
            f"""\
                    _   _            __ _     _               
                   | | | |          / _(_)   | |              
          ___  _ __| |_| |__   ___ | |_ _ ___| |__   ___ _ __ 
         / _ \| '__| __| '_ \ / _ \|  _| / __| '_ \ / _ \ '__|
        | (_) | |  | |_| | | | (_) | | | \__ \ | | |  __/ |   
         \___/|_|   \__|_| |_|\___/|_| |_|___/_| |_|\___|_|   

        Version: {__version__}
        Citation: Steenwyk et al.

        orthofisher conducts high-throughput automated sequence similarity searches across proteomes and
        creates fasta files of all significant hits. 
        
        See full documentation including a tutorial usage at the online documentation:
        https://jlsteenwyk.com/orthofisher

        Usage: orthofisher -f <list_of_fasta_files.txt> -m <list_of_hmms.txt> [optional arguments]
        """
        ),
    )

    # if no arguments are given, print help and exit
    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit()

    ## required arguments
    required = parser.add_argument_group(
        "required arguments",
        description=textwrap.dedent(
            """\
        -f, --fasta <fasta file list> 
            input file

        -m, --hmm <hmms file list>
            second input file
        """
        ),
    )

    required.add_argument(
        "-f", "--fasta", type=str, help=SUPPRESS,
    )

    required.add_argument(
        "-m", "--hmm", type=str, help=SUPPRESS,
    )

    ## optional arguments
    optional = parser.add_argument_group(
        "optional arguments",
        description=textwrap.dedent(
            """\
        -e, --evalue <e-value threshold>
            e-value threshold used when conducting sequence similarity searches

        -b, --bitscore <fraction bitscore threshold>
            fraction of a bitscore to be considered similar to top hit

        -c, --cpu <cpu workers for multithreading>
            CPU workers for multithreading

        -o, --output_dir <output directory name>
            name of the outputted directory


        -------------------------------------
        | Detailed explanation of arguments | 
        -------------------------------------
        -f, --fasta
            A one or two column tab delimited file that points to the location of fasta files that will be searched using HMMs in the first column.
            Typically, these are protein fasta files from the entire genome/transcriptome of an organism. The second column of the file specifies
            the identifier for the organism. If there is no second column, the gene identifier will be used.
        
        -m, --hmm
            A single column file with the location of HMMs that you wish to identify or fish out of a given proteome.

        -e, --evalue
            Specify an e-value threshold to use when conducting sequence similarity searches (default: 0.001). Format can be 1e-3 or 0.001.

        -b, --bitscore
            A fraction threshold to specify the bitscore threshold for sequences to be considered similar. More specifically, if a hit has a
            bitscore less than the specified fraction, the gene will not be considered sufficiently similar to be considered putatively orthologous.
            Value must range from 0 to 1 (default: 0.85).

        -c, --cpu
            Specify the number of parallel CPU workers to use for multithreading (default: 2). This argument is passed to HMMER.

        -o, --output_dir
            Name of the outputted directory with all results from the orthofisher run (default: orthofisher_output). Note, orthofisher will overwrite existing directories with the same name as specified by this argument.
        """
        ),
    )

    optional.add_argument(
        "-e",
        "--evalue",
        type=str,
        required=False,
        help=SUPPRESS,
        metavar="evalue threshold",
    )

    optional.add_argument(
        "-b",
        "--bitscore",
        type=float,
        required=False,
        help=SUPPRESS,
        metavar="bitscore threshold",
    )

    optional.add_argument(
        "-c",
        "--cpu",
        type=str,
        required=False,
        help=SUPPRESS,
        metavar="output dir",
    )

    optional.add_argument(
        "-o",
        "--output_dir",
        type=str,
        required=False,
        help=SUPPRESS,
        metavar="output dir",
    )

    optional.add_argument(
        "-h", "--help", action="help", help=SUPPRESS,
    )

    optional.add_argument(
        "-v", "--version", action="version", version=f"orthofisher {__version__}", help=SUPPRESS,
    )

    return parser


from os import path
from setuptools import setup

from orthofisher.version import __version__

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

CLASSIFIERS = [
    'Operating System :: OS Independent',
    'Intended Audience :: Science/Research',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    'Topic :: Scientific/Engineering',
]

REQUIRES = ["biopython>=1.82", "numpy>=1.24.0", "tqdm>=4.66.1", "cython"]

setup(
    name="orthofisher",
    description="orthofisher",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Jacob L. Steenwyk",
    author_email="jlsteenwyk@gmail.com",
    url="https://github.com/JLSteenwyk/orthofisher.git",
    packages=["orthofisher"],
    classifiers=CLASSIFIERS,
    entry_points={"console_scripts": ["orthofisher = orthofisher.orthofisher:main"]},
    version=__version__,
    include_package_data=True,
    install_requires=REQUIRES,
)

## push new version to pypi
# rm -rf dist
# python3 setup.py sdist bdist_wheel --universal
# twine upload dist/* -r pypi
